#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include "PsxIO.h"
#include "comm.h"

#define CHOUT	5
#define CHIN	4

/*
 * Zaehle Prozessoren. Einfach einen Count-Befehl im Kreis
 * schicken
 */
 
int countprocs (void)
{
    Comm in, out;
    int res;
    
    out.process = 0;
    out.command = COUNT;
    out.u.count = 0;
    
    res = PsxWrite (CHOUT, (char *) &out, sizeof (Comm));
    
    if (res<=0)
      return -1;
    
    res = PsxRead (CHIN, (char *) &in, sizeof (Comm));
    
    if (res<=0)
      return -1;
    
    return in.u.count;
}

/*
 * Brodcast eines Befehls. Wird verwendet fuer GO und QUIT
 */

int sendcom (int bef)
{
    Comm in, out;
    int res;
    
    out.process = 0;
    out.command = bef;
    
    res = PsxWrite (CHOUT, (char *) &out, sizeof (Comm));
    
    if (res<=0) {
	printf ("Konnte %d nicht senden! errno=%d\n", bef, errno);
	return -1;
    }
    
    res = PsxRead (CHIN, (char *) &in, sizeof (Comm));
    
    if (res<=0) {
	printf ("Konnte %d nicht empfangen! errno=%d\n", bef, errno);
	return -1;
    }

    return 0;
}

/*
 * Verschicke die Aufgaben. steps ist an dieser Stelle garantiert ein
 * positives Vielfaches der Prozessoranzahl. Jede Aufgabe wird komplett
 * durch den Ring geschickt, deshalb muessen wir den Befehl auch wieder
 * aufsammeln.
 */

int SendeAufgaben (int procs, double anfang, double ende, long steps)
{
    Comm out;
    int res, count;
    
    for (count=1; count<=procs; count++) {
	out.process        = count;
	out.command        = AUFGABE;
	out.u.aufg.anf     = anfang + (ende-anfang) * (count - 1) / procs;
	out.u.aufg.end     = anfang + (ende-anfang) * (count    ) / procs;
	out.u.aufg.schritt = (int) (steps / procs);
	
	printf ("Prozessor %2d bearbeitet Intervall %8.4lf bis %8.4lf (%d steps)\n",
		out.process, out.u.aufg.anf, out.u.aufg.end, out.u.aufg.schritt);

	res = PsxWrite (CHOUT, (char *) &out, sizeof (Comm));
	if (res!=sizeof(Comm)) {
	    printf ("Konnte Aufgabe nicht an %d schicken! errno=%d\n",
		    out.process, errno);
	    return -1;
	}
	res = PsxRead  (CHIN,  (char *) &out, sizeof (Comm));
	
	if (res!=sizeof(Comm)) {
	    printf ("Konnte Aufgabe nicht an %d schicken! errno=%d\n",
		    out.process, errno);
	    return -1;
	}
    }
    return 0;
}

/*
 * Schicke an jeden einzelnen Prozess ein ERGEBNIS-Kommando und hole
 * somit die Ergebnisse ab. Diese werden zusammenaddiert und geben
 * das Gesamtergebnis.
 */

double HoleErgebnis (int procs)
{
    Comm out, in;
    double result=0;
    int count, res;
    
    for (count=1; count<=procs; count++) {
	out.process  = count;
	out.command  = ERGEBNIS;
	
	res = PsxWrite (CHOUT, (char *) &out, sizeof (Comm));
	
	if (res<=0) {
	    printf ("Konnte Ergebnisanforderung nicht an %d schicken. errno=%d\n",
		    count, errno);
	    return 0;
	}
	
	res = PsxRead (CHIN, (char *) &in, sizeof (Comm));
	
	if (res<=0) {
	    printf ("Konnte Ergebnis von %d nicht lesen. errno=%d\n",
		    count, errno);
	    return 0;
	}
	result += in.u.ergebnis;
    }
    return result;
}

int main (int argc, char *argv[])
{
    double anfang, ende, result;
    clock_t t_begin, t_ende;	/* Zeit zur Berechnung */
    clock_t c_begin, c_count;	/* Zeit zur Kommunikation */
    long steps;
    int procs;
    
    printf ("\nMP-Integrator: sin(x)\n\n");
    printf ("Zaehle Prozessoren ...  ");
    fflush (stdout);
    
    /*
     * Zaehle Prozessoren. Dies gehoert zur Kommunikationszeit.
     */
    c_begin = clock ();
    if ((procs = countprocs ()) == -1) {
	printf ("\nFehler beim Zaehlen! errno=%d\n", errno);
	perror ("Fehler");
	exit (1);
    }
    c_count = clock () - c_begin;
    
    printf ("%d Prozessoren gefunden.\n\n", procs);
    
    /*
     * Von wo bis wo? Nehme die Werte aus Umgebungsvariablen. Wenn diese
     * nicht definiert sind, verwende die Kommandozeilenargumente (wie
     * geht das?). Falls wir auch die nicht haben, frage den User.
     */
    if (getenv ("ANFANG") && getenv ("ENDE") && getenv ("STEPS")) {
    	anfang = atof (getenv ("ANFANG"));
    	ende   = atof (getenv ("ENDE"));
    	steps  = atol (getenv ("STEPS"));
    	printf ("Anfang X .... %9.5lf\n", anfang);
    	printf ("Ende X ...... %9.5lf\n", ende);
    	printf ("Intervalle .. %9ld\n",   steps);
    }    	
    else if (argc == 4) {
    	anfang = atof (argv[1]);
    	ende   = atof (argv[2]);
    	steps  = atol (argv[3]);
    	printf ("Anfang X .... %9.5lf\n", anfang);
    	printf ("Ende X ...... %9.5lf\n", ende);
    	printf ("Intervalle .. %9ld\n",   steps);
    }
    else {
    	printf ("Anfang X .... "); fflush (stdout); scanf ("%lf", &anfang);
    	printf ("Ende X ...... "); fflush (stdout); scanf ("%lf", &ende);
    	printf ("Intervalle .. "); fflush (stdout); scanf ("%ld", &steps);
    }
    /*
     * Idiotencheck
     */
    if (steps<(long)procs) {
	printf  ("Fehler! Zu wenig Intervalle angegeben\n");
	sendcom (QUIT);
	exit (1);
    }
    /*
     * evtl. Korrektur der Schrittanzahl auf ein vielfaches der
     * Prozessoranzahl.
     */
    if (steps % procs) {
	steps += procs - (steps % procs);
	printf ("Korrigiere Intervallanzahl auf %ld\n", steps);
    }
    printf ("\nVerschicke Aufgaben ... \n");
    
    /*
     * Verschicke alle Aufgaben. Auch die hierbei verbratene Zeit wird
     * als Kommunikationszeit gewertet.
     */
    c_begin = clock ();
    if (SendeAufgaben (procs, anfang, ende, steps) != 0) {
	printf  ("Fehler beim Aufgabenverschicken!\n");
	sendcom (QUIT);
	exit (1);
    }
    c_count += clock () - c_begin;
    
    printf ("\nJetzt geht's los!\n");	
    
    /*
     * Berechnung starten! Da die ersten Prozessoren schon mit der Arbeit
     * anfangen, waehrend das Paket noch im Umlauf ist, wird auch diese
     * Zeit schon zur Rechenzeit mitgezaehlt.
     */
    t_begin = clock ();
    c_begin = clock ();
    if (sendcom (GO) != 0) {
	printf  ("Konnte Prozesse nicht starten!\n");
	sendcom (QUIT);
	exit (1);
    }
    c_count += clock () - c_begin;
    
    printf ("\nWarte auf Ergebnisse ...\n");

    /*
     * Einlesen der Ergebnisse
     */
    result = HoleErgebnis (procs);
    t_ende = clock ();
    
    if (result == 0) {
	printf  ("Konnte Ergebnis nicht lesen!\n");
	sendcom (QUIT);
	exit (1);
    }
    /*
     * Ausdruck der Zeiten und Ergebnisse
     */
    printf ("\nDas Ergebnis ist %13.10lf\n", result);
    printf (  "Verbratene Zeit: %8.5lf sec\n",
	    ((double) (t_ende - t_begin) / (double) CLK_TCK));
    printf (  "Kommunikation:   %8.5lf sec\n",
	    ((double) c_count / (double) CLK_TCK));
    
    /*
     * Schiesse slaves ab
     */
    printf ("\nBeende Prozesse\n");
    sendcom(QUIT);
    printf ("Und tschuess ...\n");
    return 0;
}
